import functools
import unittest

from fractions import Fraction

from pypol.linear import *


try:
    import sympy
    def _with_sympy(func):
        @functools.wraps(func)
        def wrapper(self):
            return func(self)
        return wrapper
except ImportError:
    def _with_sympy(func):
        @functools.wraps(func)
        def wrapper(self):
            raise unittest.SkipTest('SymPy is not available')
        return wrapper


class TestExpression(unittest.TestCase):

    def setUp(self):
        self.x = Expression({'x': 1})
        self.y = Expression({'y': 1})
        self.z = Expression({'z': 1})
        self.zero = Expression(constant=0)
        self.one = Expression(constant=1)
        self.pi = Expression(constant=Fraction(22, 7))
        self.expr = self.x - 2*self.y + 3

    def test_new_subclass(self):
        self.assertIsInstance(self.x, Symbol)
        self.assertIsInstance(self.pi, Constant)
        self.assertNotIsInstance(self.x + self.pi, Symbol)
        self.assertNotIsInstance(self.x + self.pi, Constant)
        xx = Expression({'x': 2})
        self.assertNotIsInstance(xx, Symbol)

    def test_new_types(self):
        with self.assertRaises(TypeError):
            Expression('x + y', 2)
        self.assertEqual(Expression({'x': 2}), Expression({self.x: 2}))
        with self.assertRaises(TypeError):
            Expression({0: 2})
        with self.assertRaises(TypeError):
            Expression({'x': '2'})
        self.assertEqual(Expression(constant=1), Expression(constant=self.one))
        with self.assertRaises(TypeError):
            Expression(constant='1')

    def test_symbols(self):
        self.assertCountEqual(self.x.symbols, ['x'])
        self.assertCountEqual(self.pi.symbols, [])
        self.assertCountEqual(self.expr.symbols, ['x', 'y'])

    def test_dimension(self):
        self.assertEqual(self.x.dimension, 1)
        self.assertEqual(self.pi.dimension, 0)
        self.assertEqual(self.expr.dimension, 2)

    def test_coefficient(self):
        self.assertEqual(self.expr.coefficient('x'), 1)
        self.assertEqual(self.expr.coefficient('y'), -2)
        self.assertEqual(self.expr.coefficient(self.y), -2)
        self.assertEqual(self.expr.coefficient('z'), 0)
        with self.assertRaises(TypeError):
            self.expr.coefficient(0)
        with self.assertRaises(TypeError):
            self.expr.coefficient(self.expr)

    def test_getitem(self):
        self.assertEqual(self.expr['x'], 1)
        self.assertEqual(self.expr['y'], -2)
        self.assertEqual(self.expr[self.y], -2)
        self.assertEqual(self.expr['z'], 0)
        with self.assertRaises(TypeError):
            self.expr[0]
        with self.assertRaises(TypeError):
            self.expr[self.expr]

    def test_coefficients(self):
        self.assertCountEqual(self.expr.coefficients(), [('x', 1), ('y', -2)])

    def test_constant(self):
        self.assertEqual(self.x.constant, 0)
        self.assertEqual(self.pi.constant, Fraction(22, 7))
        self.assertEqual(self.expr.constant, 3)

    def test_isconstant(self):
        self.assertFalse(self.x.isconstant())
        self.assertTrue(self.pi.isconstant())
        self.assertFalse(self.expr.isconstant())

    def test_values(self):
        self.assertCountEqual(self.expr.values(), [1, -2, 3])

    def test_issymbol(self):
        self.assertTrue(self.x.issymbol())
        self.assertFalse(self.pi.issymbol())
        self.assertFalse(self.expr.issymbol())

    def test_bool(self):
        self.assertTrue(self.x)
        self.assertFalse(self.zero)
        self.assertTrue(self.pi)
        self.assertTrue(self.expr)

    def test_pos(self):
        self.assertEqual(+self.expr, self.expr)

    def test_neg(self):
        self.assertEqual(-self.expr, -self.x + 2*self.y - 3)

    def test_add(self):
        self.assertEqual(self.x + Fraction(22, 7), self.x + self.pi)
        self.assertEqual(Fraction(22, 7) + self.x, self.x + self.pi)
        self.assertEqual(self.x + self.x, 2 * self.x)
        self.assertEqual(self.expr + 2*self.y, self.x + 3)

    def test_sub(self):
        self.assertEqual(self.x - self.x, 0)
        self.assertEqual(self.expr - 3, self.x - 2*self.y)
        self.assertEqual(0 - self.x, -self.x)

    def test_mul(self):
        self.assertEqual(self.pi * 7, 22)
        self.assertEqual(self.expr * 0, 0)
        self.assertEqual(0 * self.expr, 0)
        self.assertEqual(self.expr * 2, 2*self.x - 4*self.y + 6)

    def test_div(self):
        with self.assertRaises(ZeroDivisionError):
            self.expr / 0
        self.assertEqual(self.expr / 2, self.x / 2 - self.y + Fraction(3, 2))

    def test_str(self):
        self.assertEqual(str(Expression()), '0')
        self.assertEqual(str(self.x), 'x')
        self.assertEqual(str(-self.x), '-x')
        self.assertEqual(str(self.pi), '22/7')
        self.assertEqual(str(self.expr), 'x - 2*y + 3')

    def test_repr(self):
        self.assertEqual(repr(self.x), "Symbol('x')")
        self.assertEqual(repr(self.one), 'Constant(1)')
        self.assertEqual(repr(self.pi), 'Constant(22, 7)')
        self.assertEqual(repr(self.expr), "Expression({'x': 1, 'y': -2}, 3)")

    def test_fromstring(self):
        self.assertEqual(Expression.fromstring('x'), self.x)
        self.assertEqual(Expression.fromstring('-x'), -self.x)
        self.assertEqual(Expression.fromstring('22/7'), self.pi)
        self.assertEqual(Expression.fromstring('x - 2y + 3'), self.expr)
        self.assertEqual(Expression.fromstring('x - (3-1)y + 3'), self.expr)
        self.assertEqual(Expression.fromstring('x - 2*y + 3'), self.expr)

    def test_eq(self):
        self.assertEqual(self.expr, self.expr)
        self.assertNotEqual(self.x, self.y)
        self.assertEqual(self.zero, 0)

    def test__toint(self):
        self.assertEqual((self.x + self.y/2 + self.z/3)._toint(),
                6*self.x + 3*self.y + 2*self.z)

    @_with_sympy
    def test_fromsympy(self):
        sp_x, sp_y = sympy.symbols('x y')
        self.assertEqual(Expression.fromsympy(sp_x), self.x)
        self.assertEqual(Expression.fromsympy(sympy.Rational(22, 7)), self.pi)
        self.assertEqual(Expression.fromsympy(sp_x - 2*sp_y + 3), self.expr)
        with self.assertRaises(ValueError):
            Expression.fromsympy(sp_x*sp_y)

    @_with_sympy
    def test_tosympy(self):
        sp_x, sp_y = sympy.symbols('x y')
        self.assertEqual(self.x.tosympy(), sp_x)
        self.assertEqual(self.pi.tosympy(), sympy.Rational(22, 7))
        self.assertEqual(self.expr.tosympy(), sp_x - 2*sp_y + 3)


class TestConstant(unittest.TestCase):

    def setUp(self):
        self.zero = Constant(0)
        self.one = Constant(1)
        self.pi = Constant(Fraction(22, 7))

    @_with_sympy
    def test_fromsympy(self):
        self.assertEqual(Constant.fromsympy(sympy.Rational(22, 7)), self.pi)
        with self.assertRaises(TypeError):
            Constant.fromsympy(sympy.Symbol('x'))


class TestSymbol(unittest.TestCase):

    def setUp(self):
        self.x = Symbol('x')
        self.y = Symbol('y')

    def test_name(self):
        self.assertEqual(self.x.name, 'x')

    def test_symbols(self):
        self.assertListEqual(list(symbols('x y')), [self.x, self.y])
        self.assertListEqual(list(symbols('x,y')), [self.x, self.y])
        self.assertListEqual(list(symbols(['x', 'y'])), [self.x, self.y])

    @_with_sympy
    def test_fromsympy(self):
        sp_x = sympy.Symbol('x')
        self.assertEqual(Symbol.fromsympy(sp_x), self.x)
        with self.assertRaises(TypeError):
            Symbol.fromsympy(sympy.Rational(22, 7))
        with self.assertRaises(TypeError):
            Symbol.fromsympy(2 * sp_x)
        with self.assertRaises(TypeError):
            Symbol.fromsympy(sp_x*sp_x)


class TestOperators(unittest.TestCase):

    pass


class TestPolyhedron(unittest.TestCase):

    pass
